@@ -1,12 +1,17 @@
+use std::collections::HashMap;
+
 use {
     js_sys::WebAssembly,
     wasm_bindgen::{prelude::*, JsCast},
-    web_sys::{WebGlBuffer, WebGlProgram, WebGlRenderingContext, WebGlShader, WebGlTexture},
+    web_sys::{
+        ExtSRgb, WebGlBuffer, WebGlFramebuffer, WebGlProgram, WebGlRenderingContext, WebGlShader,
+        WebGlTexture, WebglDebugRendererInfo,
+    },
 };
 
 use egui::{
     emath::vec2,
-    epaint::{Color32, Texture},
+    epaint::{Color32, FontImage},
 };
 
 type Gl = WebGlRenderingContext;
@@ -20,23 +25,16 @@ pub struct WebGlPainter {
     pos_buffer: WebGlBuffer,
     tc_buffer: WebGlBuffer,
     color_buffer: WebGlBuffer,
+    texture_format: u32,
+    post_process: Option<PostProcess>,
 
     egui_texture: WebGlTexture,
     egui_texture_version: Option<u64>,
 
-    /// `None` means unallocated (freed) slot.
-    user_textures: Vec<Option<UserTexture>>,
-}
+    /// Index is the same as in [`egui::TextureId::User`].
+    user_textures: HashMap<u64, WebGlTexture>,
 
-#[derive(Default)]
-struct UserTexture {
-    size: (usize, usize),
-
-    /// Pending upload (will be emptied later).
-    pixels: Vec<u8>,
-
-    /// Lazily uploaded
-    gl_texture: Option<WebGlTexture>,
+    next_native_tex_id: u64, // TODO: 128-bit texture space?
 }
 
 impl WebGlPainter {
@@ -57,18 +55,36 @@ impl WebGlPainter {
         gl.tex_parameteri(Gl::TEXTURE_2D, Gl::TEXTURE_MIN_FILTER, Gl::LINEAR as i32);
         gl.tex_parameteri(Gl::TEXTURE_2D, Gl::TEXTURE_MAG_FILTER, Gl::LINEAR as i32);
 
+        let srgb_supported = matches!(gl.get_extension("EXT_sRGB"), Ok(Some(_)));
+
         let vert_shader = compile_shader(
             &gl,
             Gl::VERTEX_SHADER,
-            include_str!("shader/vertex_100es.glsl"),
-        )?;
-        let frag_shader = compile_shader(
-            &gl,
-            Gl::FRAGMENT_SHADER,
-            include_str!("shader/fragment_100es.glsl"),
+            include_str!("shader/main_vertex_100es.glsl"),
         )?;
+        let (texture_format, program, post_process) = if srgb_supported {
+            let frag_shader = compile_shader(
+                &gl,
+                Gl::FRAGMENT_SHADER,
+                include_str!("shader/main_fragment_100es.glsl"),
+            )?;
+            let program = link_program(&gl, [vert_shader, frag_shader].iter())?;
+
+            let post_process =
+                PostProcess::new(gl.clone(), canvas.width() as i32, canvas.height() as i32)?;
+
+            (ExtSRgb::SRGB_ALPHA_EXT, program, Some(post_process))
+        } else {
+            let frag_shader = compile_shader(
+                &gl,
+                Gl::FRAGMENT_SHADER,
+                include_str!("shader/fragment_100es.glsl"),
+            )?;
+            let program = link_program(&gl, [vert_shader, frag_shader].iter())?;
+
+            (Gl::RGBA, program, None)
+        };
 
-        let program = link_program(&gl, [vert_shader, frag_shader].iter())?;
         let index_buffer = gl.create_buffer().ok_or("failed to create index_buffer")?;
         let pos_buffer = gl.create_buffer().ok_or("failed to create pos_buffer")?;
         let tc_buffer = gl.create_buffer().ok_or("failed to create tc_buffer")?;
@@ -83,123 +99,22 @@ impl WebGlPainter {
             pos_buffer,
             tc_buffer,
             color_buffer,
+            texture_format,
+            post_process,
             egui_texture,
             egui_texture_version: None,
             user_textures: Default::default(),
+            next_native_tex_id: 1 << 32,
         })
     }
 
-    fn alloc_user_texture_index(&mut self) -> usize {
-        for (index, tex) in self.user_textures.iter_mut().enumerate() {
-            if tex.is_none() {
-                *tex = Some(Default::default());
-                return index;
-            }
-        }
-        let index = self.user_textures.len();
-        self.user_textures.push(Some(Default::default()));
-        index
-    }
-
-    fn alloc_user_texture(
-        &mut self,
-        size: (usize, usize),
-        srgba_pixels: &[Color32],
-    ) -> egui::TextureId {
-        let index = self.alloc_user_texture_index();
-        assert_eq!(size.0 * size.1, srgba_pixels.len());
-
-        if let Some(Some(user_texture)) = self.user_textures.get_mut(index) {
-            let mut pixels: Vec<u8> = Vec::with_capacity(srgba_pixels.len() * 4);
-            for srgba in srgba_pixels {
-                pixels.push(srgba.r());
-                pixels.push(srgba.g());
-                pixels.push(srgba.b());
-                pixels.push(srgba.a());
-            }
-
-            *user_texture = UserTexture {
-                size,
-                pixels,
-                gl_texture: None,
-            };
-        }
-
-        egui::TextureId::User(index as u64)
-    }
-
-    fn free_user_texture(&mut self, id: egui::TextureId) {
-        if let egui::TextureId::User(id) = id {
-            let index = id as usize;
-            if index < self.user_textures.len() {
-                self.user_textures[index] = None;
-            }
-        }
-    }
-    pub fn get_texture(&self, texture_id: egui::TextureId) -> Option<&WebGlTexture> {
+    fn get_texture(&self, texture_id: egui::TextureId) -> Option<&WebGlTexture> {
         match texture_id {
             egui::TextureId::Egui => Some(&self.egui_texture),
-            egui::TextureId::User(id) => self
-                .user_textures
-                .get(id as usize)?
-                .as_ref()?
-                .gl_texture
-                .as_ref(),
+            egui::TextureId::User(id) => self.user_textures.get(&id),
         }
     }
 
-    fn upload_user_textures(&mut self) {
-        let gl = &self.gl;
-
-        for user_texture in self.user_textures.iter_mut().flatten() {
-            if user_texture.gl_texture.is_none() {
-                let pixels = std::mem::take(&mut user_texture.pixels);
-
-                let gl_texture = gl.create_texture().unwrap();
-                gl.bind_texture(Gl::TEXTURE_2D, Some(&gl_texture));
-                gl.tex_parameteri(Gl::TEXTURE_2D, Gl::TEXTURE_WRAP_S, Gl::CLAMP_TO_EDGE as i32);
-                gl.tex_parameteri(Gl::TEXTURE_2D, Gl::TEXTURE_WRAP_T, Gl::CLAMP_TO_EDGE as i32);
-                gl.tex_parameteri(Gl::TEXTURE_2D, Gl::TEXTURE_MIN_FILTER, Gl::LINEAR as i32);
-                gl.tex_parameteri(Gl::TEXTURE_2D, Gl::TEXTURE_MAG_FILTER, Gl::LINEAR as i32);
-
-                gl.bind_texture(Gl::TEXTURE_2D, Some(&gl_texture));
-
-                // TODO: https://developer.mozilla.org/en-US/docs/Web/API/EXT_sRGB
-                let level = 0;
-                let internal_format = Gl::RGBA;
-                let border = 0;
-                let src_format = Gl::RGBA;
-                let src_type = Gl::UNSIGNED_BYTE;
-                gl.tex_image_2d_with_i32_and_i32_and_i32_and_format_and_type_and_opt_u8_array(
-                    Gl::TEXTURE_2D,
-                    level,
-                    internal_format as i32,
-                    user_texture.size.0 as i32,
-                    user_texture.size.1 as i32,
-                    border,
-                    src_format,
-                    src_type,
-                    Some(&pixels),
-                )
-                .unwrap();
-
-                user_texture.gl_texture = Some(gl_texture);
-            }
-        }
-    }
-
-    pub fn register_webgl_texture(&mut self, texture: WebGlTexture) -> egui::TextureId {
-        let id = self.alloc_user_texture_index();
-        if let Some(Some(user_texture)) = self.user_textures.get_mut(id) {
-            *user_texture = UserTexture {
-                size: (0, 0),
-                pixels: vec![],
-                gl_texture: Some(texture),
-            }
-        }
-        egui::TextureId::User(id as u64)
-    }
-
     fn paint_mesh(&self, mesh: &egui::epaint::Mesh16) -> Result<(), JsValue> {
         debug_assert!(mesh.is_valid());
 
@@ -321,23 +236,75 @@ impl WebGlPainter {
     }
 }
 
-impl epi::TextureAllocator for WebGlPainter {
-    fn alloc_srgba_premultiplied(
-        &mut self,
-        size: (usize, usize),
-        srgba_pixels: &[egui::Color32],
-    ) -> egui::TextureId {
-        self.alloc_user_texture(size, srgba_pixels)
+impl epi::NativeTexture for WebGlPainter {
+    type Texture = WebGlTexture;
+
+    fn register_native_texture(&mut self, native: Self::Texture) -> egui::TextureId {
+        let id = self.next_native_tex_id;
+        self.next_native_tex_id += 1;
+        self.user_textures.insert(id, native);
+        egui::TextureId::User(id as u64)
     }
 
-    fn free(&mut self, id: egui::TextureId) {
-        self.free_user_texture(id)
+    fn replace_native_texture(&mut self, id: egui::TextureId, replacing: Self::Texture) {
+        if let egui::TextureId::User(id) = id {
+            if let Some(user_texture) = self.user_textures.get_mut(&id) {
+                *user_texture = replacing;
+            }
+        }
     }
 }
 
 impl crate::Painter for WebGlPainter {
-    fn as_tex_allocator(&mut self) -> &mut dyn epi::TextureAllocator {
-        self
+    fn set_texture(&mut self, tex_id: u64, image: epi::Image) {
+        assert_eq!(
+            image.size[0] * image.size[1],
+            image.pixels.len(),
+            "Mismatch between texture size and texel count"
+        );
+
+        // TODO: optimize
+        let mut pixels: Vec<u8> = Vec::with_capacity(image.pixels.len() * 4);
+        for srgba in image.pixels {
+            pixels.push(srgba.r());
+            pixels.push(srgba.g());
+            pixels.push(srgba.b());
+            pixels.push(srgba.a());
+        }
+
+        let gl = &self.gl;
+        let gl_texture = gl.create_texture().unwrap();
+        gl.bind_texture(Gl::TEXTURE_2D, Some(&gl_texture));
+        gl.tex_parameteri(Gl::TEXTURE_2D, Gl::TEXTURE_WRAP_S, Gl::CLAMP_TO_EDGE as _);
+        gl.tex_parameteri(Gl::TEXTURE_2D, Gl::TEXTURE_WRAP_T, Gl::CLAMP_TO_EDGE as _);
+        gl.tex_parameteri(Gl::TEXTURE_2D, Gl::TEXTURE_MIN_FILTER, Gl::LINEAR as _);
+        gl.tex_parameteri(Gl::TEXTURE_2D, Gl::TEXTURE_MAG_FILTER, Gl::LINEAR as _);
+
+        gl.bind_texture(Gl::TEXTURE_2D, Some(&gl_texture));
+
+        let level = 0;
+        let internal_format = self.texture_format;
+        let border = 0;
+        let src_format = self.texture_format;
+        let src_type = Gl::UNSIGNED_BYTE;
+        gl.tex_image_2d_with_i32_and_i32_and_i32_and_format_and_type_and_opt_u8_array(
+            Gl::TEXTURE_2D,
+            level,
+            internal_format as _,
+            image.size[0] as _,
+            image.size[1] as _,
+            border,
+            src_format,
+            src_type,
+            Some(&pixels),
+        )
+        .unwrap();
+
+        self.user_textures.insert(tex_id, gl_texture);
+    }
+
+    fn free_texture(&mut self, tex_id: u64) {
+        self.user_textures.remove(&tex_id);
     }
 
     fn debug_info(&self) -> String {
@@ -356,13 +323,18 @@ impl crate::Painter for WebGlPainter {
         &self.canvas_id
     }
 
-    fn upload_egui_texture(&mut self, texture: &Texture) {
-        if self.egui_texture_version == Some(texture.version) {
+    fn upload_egui_texture(&mut self, font_image: &FontImage) {
+        if self.egui_texture_version == Some(font_image.version) {
             return; // No change
         }
 
-        let mut pixels: Vec<u8> = Vec::with_capacity(texture.pixels.len() * 4);
-        for srgba in texture.srgba_pixels() {
+        let gamma = if self.post_process.is_none() {
+            1.0 / 2.2 // HACK due to non-linear framebuffer blending.
+        } else {
+            1.0 // post process enables linear blending
+        };
+        let mut pixels: Vec<u8> = Vec::with_capacity(font_image.pixels.len() * 4);
+        for srgba in font_image.srgba_pixels(gamma) {
             pixels.push(srgba.r());
             pixels.push(srgba.g());
             pixels.push(srgba.b());
@@ -372,19 +344,17 @@ impl crate::Painter for WebGlPainter {
         let gl = &self.gl;
         gl.bind_texture(Gl::TEXTURE_2D, Some(&self.egui_texture));
 
-        // TODO: https://developer.mozilla.org/en-US/docs/Web/API/EXT_sRGB
-        // https://www.khronos.org/registry/webgl/extensions/EXT_sRGB/
         let level = 0;
-        let internal_format = Gl::RGBA;
+        let internal_format = self.texture_format;
         let border = 0;
-        let src_format = Gl::RGBA;
+        let src_format = self.texture_format;
         let src_type = Gl::UNSIGNED_BYTE;
         gl.tex_image_2d_with_i32_and_i32_and_i32_and_format_and_type_and_opt_u8_array(
             Gl::TEXTURE_2D,
             level,
             internal_format as i32,
-            texture.width as i32,
-            texture.height as i32,
+            font_image.width as i32,
+            font_image.height as i32,
             border,
             src_format,
             src_type,
@@ -392,7 +362,7 @@ impl crate::Painter for WebGlPainter {
         )
         .unwrap();
 
-        self.egui_texture_version = Some(texture.version);
+        self.egui_texture_version = Some(font_image.version);
     }
 
     fn clear(&mut self, clear_color: egui::Rgba) {
@@ -419,10 +389,12 @@ impl crate::Painter for WebGlPainter {
         clipped_meshes: Vec<egui::ClippedMesh>,
         pixels_per_point: f32,
     ) -> Result<(), JsValue> {
-        self.upload_user_textures();
-
         let gl = &self.gl;
 
+        if let Some(ref mut post_process) = self.post_process {
+            post_process.begin(self.canvas.width() as i32, self.canvas.height() as i32)?;
+        }
+
         gl.enable(Gl::SCISSOR_TEST);
         gl.disable(Gl::CULL_FACE); // egui is not strict about winding order.
         gl.enable(Gl::BLEND);
@@ -479,8 +451,193 @@ impl crate::Painter for WebGlPainter {
                 ));
             }
         }
+
+        if let Some(ref post_process) = self.post_process {
+            post_process.end();
+        }
+
         Ok(())
     }
+
+    fn name(&self) -> &'static str {
+        "egui_web (WebGL1)"
+    }
+}
+
+struct PostProcess {
+    gl: Gl,
+    pos_buffer: WebGlBuffer,
+    a_pos_loc: u32,
+    index_buffer: WebGlBuffer,
+    texture: WebGlTexture,
+    texture_size: (i32, i32),
+    fbo: WebGlFramebuffer,
+    program: WebGlProgram,
+}
+
+fn requires_brightening(gl: &web_sys::WebGlRenderingContext) -> bool {
+    // See https://github.com/emilk/egui/issues/794
+
+    let user_agent = web_sys::window().unwrap().navigator().user_agent().unwrap();
+    crate::webgl1::is_safari_and_webkit_gtk(gl) && !user_agent.contains("Mac OS X")
+}
+
+impl PostProcess {
+    fn new(gl: Gl, width: i32, height: i32) -> Result<PostProcess, JsValue> {
+        let fbo = gl
+            .create_framebuffer()
+            .ok_or("failed to create framebuffer")?;
+        gl.bind_framebuffer(Gl::FRAMEBUFFER, Some(&fbo));
+
+        let texture = gl.create_texture().unwrap();
+        gl.bind_texture(Gl::TEXTURE_2D, Some(&texture));
+        gl.tex_parameteri(Gl::TEXTURE_2D, Gl::TEXTURE_WRAP_S, Gl::CLAMP_TO_EDGE as i32);
+        gl.tex_parameteri(Gl::TEXTURE_2D, Gl::TEXTURE_WRAP_T, Gl::CLAMP_TO_EDGE as i32);
+        gl.tex_parameteri(Gl::TEXTURE_2D, Gl::TEXTURE_MIN_FILTER, Gl::NEAREST as i32);
+        gl.tex_parameteri(Gl::TEXTURE_2D, Gl::TEXTURE_MAG_FILTER, Gl::NEAREST as i32);
+        gl.pixel_storei(Gl::UNPACK_ALIGNMENT, 1);
+        gl.tex_image_2d_with_i32_and_i32_and_i32_and_format_and_type_and_opt_u8_array(
+            Gl::TEXTURE_2D,
+            0,
+            ExtSRgb::SRGB_ALPHA_EXT as i32,
+            width,
+            height,
+            0,
+            ExtSRgb::SRGB_ALPHA_EXT,
+            Gl::UNSIGNED_BYTE,
+            None,
+        )
+        .unwrap();
+        gl.framebuffer_texture_2d(
+            Gl::FRAMEBUFFER,
+            Gl::COLOR_ATTACHMENT0,
+            Gl::TEXTURE_2D,
+            Some(&texture),
+            0,
+        );
+
+        gl.bind_texture(Gl::TEXTURE_2D, None);
+        gl.bind_framebuffer(Gl::FRAMEBUFFER, None);
+
+        let shader_prefix = if requires_brightening(&gl) {
+            crate::console_log("Enabling webkitGTK brightening workaround");
+            "#define APPLY_BRIGHTENING_GAMMA"
+        } else {
+            ""
+        };
+
+        let vert_shader = compile_shader(
+            &gl,
+            Gl::VERTEX_SHADER,
+            include_str!("shader/post_vertex_100es.glsl"),
+        )?;
+        let frag_shader = compile_shader(
+            &gl,
+            Gl::FRAGMENT_SHADER,
+            &format!(
+                "{}{}",
+                shader_prefix,
+                include_str!("shader/post_fragment_100es.glsl")
+            ),
+        )?;
+        let program = link_program(&gl, [vert_shader, frag_shader].iter())?;
+
+        let positions = vec![0u8, 0, 1, 0, 0, 1, 1, 1];
+
+        let indices = vec![0u8, 1, 2, 1, 2, 3];
+
+        let pos_buffer = gl.create_buffer().ok_or("failed to create pos_buffer")?;
+        gl.bind_buffer(Gl::ARRAY_BUFFER, Some(&pos_buffer));
+        gl.buffer_data_with_u8_array(Gl::ARRAY_BUFFER, &positions, Gl::STATIC_DRAW);
+        gl.bind_buffer(Gl::ARRAY_BUFFER, None);
+
+        let a_pos_loc = gl.get_attrib_location(&program, "a_pos");
+        assert!(a_pos_loc >= 0);
+        let a_pos_loc = a_pos_loc as u32;
+
+        let index_buffer = gl.create_buffer().ok_or("failed to create index_buffer")?;
+        gl.bind_buffer(Gl::ELEMENT_ARRAY_BUFFER, Some(&index_buffer));
+        gl.buffer_data_with_u8_array(Gl::ELEMENT_ARRAY_BUFFER, &indices, Gl::STATIC_DRAW);
+        gl.bind_buffer(Gl::ELEMENT_ARRAY_BUFFER, None);
+
+        Ok(PostProcess {
+            gl,
+            pos_buffer,
+            a_pos_loc,
+            index_buffer,
+            texture,
+            texture_size: (width, height),
+            fbo,
+            program,
+        })
+    }
+
+    fn begin(&mut self, width: i32, height: i32) -> Result<(), JsValue> {
+        let gl = &self.gl;
+
+        if (width, height) != self.texture_size {
+            gl.bind_texture(Gl::TEXTURE_2D, Some(&self.texture));
+            gl.pixel_storei(Gl::UNPACK_ALIGNMENT, 1);
+            gl.tex_image_2d_with_i32_and_i32_and_i32_and_format_and_type_and_opt_u8_array(
+                Gl::TEXTURE_2D,
+                0,
+                ExtSRgb::SRGB_ALPHA_EXT as i32,
+                width,
+                height,
+                0,
+                ExtSRgb::SRGB_ALPHA_EXT,
+                Gl::UNSIGNED_BYTE,
+                None,
+            )?;
+            gl.bind_texture(Gl::TEXTURE_2D, None);
+
+            self.texture_size = (width, height);
+        }
+
+        gl.bind_framebuffer(Gl::FRAMEBUFFER, Some(&self.fbo));
+        gl.clear_color(0.0, 0.0, 0.0, 0.0);
+        gl.clear(Gl::COLOR_BUFFER_BIT);
+
+        Ok(())
+    }
+
+    fn end(&self) {
+        let gl = &self.gl;
+
+        gl.bind_framebuffer(Gl::FRAMEBUFFER, None);
+        gl.disable(Gl::SCISSOR_TEST);
+
+        gl.use_program(Some(&self.program));
+
+        gl.active_texture(Gl::TEXTURE0);
+        gl.bind_texture(Gl::TEXTURE_2D, Some(&self.texture));
+        let u_sampler_loc = gl.get_uniform_location(&self.program, "u_sampler").unwrap();
+        gl.uniform1i(Some(&u_sampler_loc), 0);
+
+        gl.bind_buffer(Gl::ARRAY_BUFFER, Some(&self.pos_buffer));
+        gl.vertex_attrib_pointer_with_i32(self.a_pos_loc, 2, Gl::UNSIGNED_BYTE, false, 0, 0);
+        gl.enable_vertex_attrib_array(self.a_pos_loc);
+
+        gl.bind_buffer(Gl::ELEMENT_ARRAY_BUFFER, Some(&self.index_buffer));
+
+        gl.draw_elements_with_i32(Gl::TRIANGLES, 6, Gl::UNSIGNED_BYTE, 0);
+
+        gl.bind_buffer(Gl::ELEMENT_ARRAY_BUFFER, None);
+        gl.bind_buffer(Gl::ARRAY_BUFFER, None);
+        gl.bind_texture(Gl::TEXTURE_2D, None);
+        gl.use_program(None);
+    }
+}
+
+impl Drop for PostProcess {
+    fn drop(&mut self) {
+        let gl = &self.gl;
+        gl.delete_buffer(Some(&self.pos_buffer));
+        gl.delete_buffer(Some(&self.index_buffer));
+        gl.delete_program(Some(&self.program));
+        gl.delete_framebuffer(Some(&self.fbo));
+        gl.delete_texture(Some(&self.texture));
+    }
 }
 
 fn compile_shader(
@@ -531,3 +688,26 @@ fn link_program<'a, T: IntoIterator<Item = &'a WebGlShader>>(
             .unwrap_or_else(|| "Unknown error creating program object".into()))
     }
 }
+
+/// detecting Safari and webkitGTK.
+///
+/// Safari and webkitGTK use unmasked renderer :Apple GPU
+///
+/// If we detect safari or webkitGTK returns true.
+///
+/// This function used to avoid displaying linear color with `sRGB` supported systems.
+pub(crate) fn is_safari_and_webkit_gtk(gl: &web_sys::WebGlRenderingContext) -> bool {
+    if gl
+        .get_extension("WEBGL_debug_renderer_info")
+        .unwrap()
+        .is_some()
+    {
+        let renderer: JsValue = gl
+            .get_parameter(WebglDebugRendererInfo::UNMASKED_RENDERER_WEBGL)
+            .unwrap();
+        if renderer.as_string().unwrap().contains("Apple") {
+            return true;
+        }
+    }
+    false
+}
